@@ -1,13 +1,19 @@
 import datetime
+import re
+from weakref import WeakKeyDictionary
 
-import six
 from dateutil.parser import parse
 
 from .errors import ValidationError
 from .collections import ModelCollection
 
 
-class BaseField(object):
+# unique marker for "no default value specified". None is not good enough since
+# it is a completely valid default value.
+NotSet = object()
+
+
+class BaseField:
 
     """Base class for all fields."""
 
@@ -16,32 +22,52 @@ class BaseField(object):
     def __init__(
             self,
             required=False,
+            nullable=False,
             help_text=None,
-            validators=None):
-        self._memory = {}
+            validators=None,
+            default=NotSet,
+            name=None):
+        self.memory = WeakKeyDictionary()
         self.required = required
         self.help_text = help_text
+        self.nullable = nullable
         self._assign_validators(validators)
+        self.name = name
+        self._validate_name()
+        if default is not NotSet:
+            self.validate(default)
+        self._default = default
+
+    @property
+    def has_default(self):
+        return self._default is not NotSet
 
     def _assign_validators(self, validators):
         if validators and not isinstance(validators, list):
             validators = [validators]
         self.validators = validators or []
 
-    def __set__(self, obj, value):
+    def __set__(self, instance, value):
+        self._finish_initialization(type(instance))
         value = self.parse_value(value)
         self.validate(value)
-        self._memory[obj] = value
+        self.memory[instance._cache_key] = value
 
-    def __get__(self, obj, owner=None):
-        if obj is None:
+    def __get__(self, instance, owner=None):
+        if instance is None:
+            self._finish_initialization(owner)
             return self
 
-        self._check_value(obj)
-        return self._memory[obj]
+        self._finish_initialization(type(instance))
+
+        self._check_value(instance)
+        return self.memory[instance._cache_key]
+
+    def _finish_initialization(self, owner):
+        pass
 
     def _check_value(self, obj):
-        if obj not in self._memory:
+        if obj._cache_key not in self.memory:
             self.__set__(obj, self.get_default_value())
 
     def validate_for_object(self, obj):
@@ -61,17 +87,17 @@ class BaseField(object):
     def _validate_against_types(self, value):
         if value is not None and not isinstance(value, self.types):
             raise ValidationError(
-                'Value is wrong, expected type "{}"'.format(
-                    ', '.join([t.__name__ for t in self.types])
+                'Value is wrong, expected type "{types}"'.format(
+                    types=', '.join([t.__name__ for t in self.types])
                 ),
-                value
+                value,
             )
 
     def _check_types(self):
         if self.types is None:
             raise ValidationError(
-                'Field "{}" is not usable, try '
-                'different field type.'.format(type(self).__name__))
+                'Field "{type}" is not usable, try '
+                'different field type.'.format(type=type(self).__name__))
 
     def to_struct(self, value):
         """Cast value to Python structure."""
@@ -87,27 +113,38 @@ class BaseField(object):
         return value
 
     def _validate_with_custom_validators(self, value):
+        if value is None and self.nullable:
+            return
+
         for validator in self.validators:
             try:
                 validator.validate(value)
             except AttributeError:
                 validator(value)
 
-    @staticmethod
-    def get_default_value():
+    def get_default_value(self):
         """Get default value for field.
 
         Each field can specify its default.
 
         """
-        return None
+        return self._default if self.has_default else None
+
+    def _validate_name(self):
+        if self.name is None:
+            return
+        if not re.match(r'^[A-Za-z_](([\w\-]*)?\w+)?$', self.name):
+            raise ValueError('Wrong name', self.name)
+
+    def structue_name(self, default):
+        return self.name if self.name is not None else default
 
 
 class StringField(BaseField):
 
     """String field."""
 
-    types = six.string_types
+    types = (str,)
 
 
 class IntField(BaseField):
@@ -116,6 +153,13 @@ class IntField(BaseField):
 
     types = (int,)
 
+    def parse_value(self, value):
+        """Cast value to `int`, e.g. from string or long"""
+        parsed = super().parse_value(value)
+        if parsed is None:
+            return parsed
+        return int(parsed)
+
 
 class FloatField(BaseField):
 
@@ -132,7 +176,7 @@ class BoolField(BaseField):
 
     def parse_value(self, value):
         """Cast value to `bool`."""
-        parsed = super(BoolField, self).parse_value(value)
+        parsed = super().parse_value(value)
         return bool(parsed) if parsed is not None else None
 
 
@@ -150,9 +194,15 @@ class ListField(BaseField):
 
         """
         self._assign_types(items_types)
-        super(ListField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
         self.required = False
 
+    def get_default_value(self):
+        default = super().get_default_value()
+        if default is None:
+            return ModelCollection(self)
+        return default
+
     def _assign_types(self, items_types):
         if items_types:
             try:
@@ -162,17 +212,22 @@ class ListField(BaseField):
         else:
             self.items_types = tuple()
 
+        types = []
+        for type_ in self.items_types:
+            if isinstance(type_, str):
+                types.append(_LazyType(type_))
+            else:
+                types.append(type_)
+        self.items_types = tuple(types)
+
     def validate(self, value):
-        super(ListField, self).validate(value)
+        super().validate(value)
 
         if len(self.items_types) == 0:
             return
 
-        try:
-            for item in value:
-                self.validate_single_value(item)
-        except TypeError:
-            pass
+        for item in value:
+            self.validate_single_value(item)
 
     def validate_single_value(self, item):
         if len(self.items_types) == 0:
@@ -181,18 +236,11 @@ class ListField(BaseField):
         if not isinstance(item, self.items_types):
             raise ValidationError(
                 'All items must be instances '
-                'of "{}", and not "{}".'.format(
-                    ', '.join([t.__name__ for t in self.items_types]),
-                    type(item).__name__
+                'of "{types}", and not "{type}".'.format(
+                    types=', '.join([t.__name__ for t in self.items_types]),
+                    type=type(item).__name__,
                 ))
 
-    def to_struct(self, value):
-        """Cast value to list."""
-        return [item.to_struct() for item in value]
-
-    def get_default_value(self):
-        return ModelCollection(self)
-
     def parse_value(self, values):
         """Cast value to proper collection."""
         result = self.get_default_value()
@@ -203,36 +251,40 @@ class ListField(BaseField):
         if not isinstance(values, list):
             return values
 
-        embed_type = self.items_types[0]
+        return [self._cast_value(value) for value in values]
 
-        if not hasattr(getattr(embed_type, 'populate', None), '__call__'):
-            return values
-
-        self._check_items_types_count()
-        self._parse_values_to_result(values, embed_type, result)
-
-        return result
-
-    def _check_items_types_count(self):
-        if len(self.items_types) != 1:
-            raise ValidationError(
-                'Cannot decide which type to choose from "{}".'.format(
-                    ', '.join([t.__name__ for t in self.items_types])
-                )
-            )
-
-    def _parse_values_to_result(self, values, embed_type, result):
-        try:
-            for value in values:
-                self._append_value_to_result(value, result, embed_type)
-        except TypeError:
-            raise ValidationError('Given value for field is not iterable.')
-
-    def _append_value_to_result(self, value, result, embed_type):
+    def _cast_value(self, value):
         if isinstance(value, self.items_types):
-            result.append(value)
+            return value
         else:
-            result.append(embed_type(**value))
+            if len(self.items_types) != 1:
+                tpl = 'Cannot decide which type to choose from "{types}".'
+                raise ValidationError(
+                    tpl.format(
+                        types=', '.join([t.__name__ for t in self.items_types])
+                    )
+                )
+            return self.items_types[0](**value)
+
+    def _finish_initialization(self, owner):
+        super()._finish_initialization(owner)
+
+        types = []
+        for type in self.items_types:
+            if isinstance(type, _LazyType):
+                types.append(type.evaluate(owner))
+            else:
+                types.append(type)
+        self.items_types = tuple(types)
+
+    def _elem_to_struct(self, value):
+        try:
+            return value.to_struct()
+        except AttributeError:
+            return value
+
+    def to_struct(self, values):
+        return [self._elem_to_struct(v) for v in values]
 
 
 class EmbeddedField(BaseField):
@@ -241,17 +293,33 @@ class EmbeddedField(BaseField):
 
     def __init__(self, model_types, *args, **kwargs):
         self._assign_model_types(model_types)
-        super(EmbeddedField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def _assign_model_types(self, model_types):
-        try:
-            iter(model_types)
-            self.types = tuple(model_types)
-        except TypeError:
-            self.types = (model_types,)
+        if not isinstance(model_types, (list, tuple)):
+            model_types = (model_types,)
+
+        types = []
+        for type_ in model_types:
+            if isinstance(type_, str):
+                types.append(_LazyType(type_))
+            else:
+                types.append(type_)
+        self.types = tuple(types)
+
+    def _finish_initialization(self, owner):
+        super()._finish_initialization(owner)
+
+        types = []
+        for type in self.types:
+            if isinstance(type, _LazyType):
+                types.append(type.evaluate(owner))
+            else:
+                types.append(type)
+        self.types = tuple(types)
 
     def validate(self, value):
-        super(EmbeddedField, self).validate(value)
+        super().validate(value)
         try:
             value.validate()
         except AttributeError:
@@ -268,12 +336,61 @@ class EmbeddedField(BaseField):
     def _get_embed_type(self):
         if len(self.types) != 1:
             raise ValidationError(
-                'Cannot decide which type to choose from "{}".'.format(
-                    ', '.join([t.__name__ for t in self.types])
+                'Cannot decide which type to choose from "{types}".'.format(
+                    types=', '.join([t.__name__ for t in self.types])
                 )
             )
         return self.types[0]
 
+    def to_struct(self, value):
+        return value.to_struct()
+
+
+class _LazyType:
+
+    def __init__(self, path):
+        self.path = path
+
+    def evaluate(self, base_cls):
+        module, type_name = _evaluate_path(self.path, base_cls)
+        return _import(module, type_name)
+
+
+def _evaluate_path(relative_path, base_cls):
+    base_module = base_cls.__module__
+
+    modules = _get_modules(relative_path, base_module)
+
+    type_name = modules.pop()
+    module = '.'.join(modules)
+    if not module:
+        module = base_module
+    return module, type_name
+
+
+def _get_modules(relative_path, base_module):
+    canonical_path = relative_path.lstrip('.')
+    canonical_modules = canonical_path.split('.')
+
+    if not relative_path.startswith('.'):
+        return canonical_modules
+
+    parents_amount = len(relative_path) - len(canonical_path)
+    parent_modules = base_module.split('.')
+    parents_amount = max(0, parents_amount - 1)
+    if parents_amount > len(parent_modules):
+        raise ValueError(f"Can't evaluate path '{relative_path}'")
+    return parent_modules[:parents_amount * -1] + canonical_modules
+
+
+def _import(module_name, type_name):
+    module = __import__(module_name, fromlist=[type_name])
+    try:
+        return getattr(module, type_name)
+    except AttributeError:
+        raise ValueError(
+            f"Can't find type '{module_name}.{type_name}'.")
+
 
 class TimeField(StringField):
 
@@ -289,7 +406,7 @@ class TimeField(StringField):
 
         """
         self.str_format = str_format
-        super(TimeField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def to_struct(self, value):
         """Cast `time` object to string."""
@@ -299,6 +416,8 @@ class TimeField(StringField):
 
     def parse_value(self, value):
         """Parse string into instance of `time`."""
+        if value is None:
+            return value
         if isinstance(value, datetime.time):
             return value
         return parse(value).timetz()
@@ -319,7 +438,7 @@ class DateField(StringField):
 
         """
         self.str_format = str_format
-        super(DateField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def to_struct(self, value):
         """Cast `date` object to string."""
@@ -329,6 +448,8 @@ class DateField(StringField):
 
     def parse_value(self, value):
         """Parse string into instance of `date`."""
+        if value is None:
+            return value
         if isinstance(value, datetime.date):
             return value
         return parse(value).date()
@@ -348,7 +469,7 @@ class DateTimeField(StringField):
 
         """
         self.str_format = str_format
-        super(DateTimeField, self).__init__(*args, **kwargs)
+        super().__init__(*args, **kwargs)
 
     def to_struct(self, value):
         """Cast `datetime` object to string."""
@@ -360,4 +481,7 @@ class DateTimeField(StringField):
         """Parse string into instance of `datetime`."""
         if isinstance(value, datetime.datetime):
             return value
-        return parse(value)
+        if value:
+            return parse(value)
+        else:
+            return None
